"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.deobfuscate = void 0;
const shift_parser_1 = require("shift-parser");
const shift_codegen_1 = require("shift-codegen");
const proxyRemover_1 = __importDefault(require("./modifications/proxies/proxyRemover"));
const expressionSimplifier_1 = __importDefault(require("./modifications/expressions/expressionSimplifier"));
const arrayUnpacker_1 = __importDefault(require("./modifications/arrays/arrayUnpacker"));
const propertySimplifier_1 = __importDefault(require("./modifications/properties/propertySimplifier"));
const cleanupHelper_1 = __importDefault(require("./helpers/cleanupHelper"));
const variableRenamer_1 = __importDefault(require("./modifications/renaming/variableRenamer"));
const deadBranchRemover_1 = __importDefault(require("./modifications/branches/deadBranchRemover"));
const stringDecoder_1 = __importDefault(require("./modifications/expressions/stringDecoder"));
const defaultConfig = {
    verbose: false,
    isModule: false,
    arrays: {
        unpackArrays: true,
        removeArrays: true
    },
    proxyFunctions: {
        replaceProxyFunctions: true,
        removeProxyFunctions: true
    },
    expressions: {
        simplifyExpressions: true,
        removeDeadBranches: true,
        undoStringOperations: true
    },
    miscellaneous: {
        beautify: true,
        simplifyProperties: true,
        renameHexIdentifiers: false
    }
};
/**
 * Deobfuscates a given source script.
 * @param source The source script.
 * @param config The deobfuscation configuration (optional).
 * @returns The deobfuscated script.
 */
function deobfuscate(source, parsedConfig) {
    const config = Object.assign({}, defaultConfig, parsedConfig);
    const ast = (config.isModule ? (0, shift_parser_1.parseModule)(source) : (0, shift_parser_1.parseScript)(source));
    const modifications = [];
    if (config.proxyFunctions.replaceProxyFunctions) {
        modifications.push(new proxyRemover_1.default(ast, config.proxyFunctions.removeProxyFunctions));
    }
    if (config.expressions.simplifyExpressions) {
        modifications.push(new expressionSimplifier_1.default(ast));
    }
    if (config.arrays.unpackArrays) {
        modifications.push(new arrayUnpacker_1.default(ast, config.arrays.removeArrays));
    }
    // simplify any expressions that were revealed by the array unpacking
    if (config.expressions.simplifyExpressions) {
        modifications.push(new expressionSimplifier_1.default(ast));
    }
    if (config.expressions.removeDeadBranches) {
        modifications.push(new deadBranchRemover_1.default(ast));
    }
    if (config.miscellaneous.simplifyProperties) {
        modifications.push(new propertySimplifier_1.default(ast));
    }
    if (config.expressions.undoStringOperations) {
        modifications.push(new stringDecoder_1.default(ast));
    }
    if (config.miscellaneous.renameHexIdentifiers) {
        modifications.push(new variableRenamer_1.default(ast));
    }
    for (const modification of modifications) {
        if (config.verbose) {
            console.log(`[${new Date().toISOString()}]: Executing ${modification.constructor.name}`);
        }
        modification.execute();
    }
    cleanupHelper_1.default.cleanup(ast);
    const output = config.miscellaneous.beautify
        ? (0, shift_codegen_1.codeGen)(ast, new shift_codegen_1.FormattedCodeGen())
        : (0, shift_codegen_1.codeGen)(ast);
    return output;
}
exports.deobfuscate = deobfuscate;
