"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const modification_1 = __importDefault(require("../../modification"));
const Shift = __importStar(require("shift-ast"));
const traverse_1 = require("../../helpers/traverse");
const traversalHelper_1 = __importDefault(require("../../helpers/traversalHelper"));
class ExpressionSimplifier extends modification_1.default {
    /**
     * Creates a new modification.
     * @param ast The AST.
     */
    constructor(ast) {
        super('Simplify Expressions', ast);
        this.types = new Set(['BinaryExpression', 'UnaryExpression']);
    }
    /**
     * Executes the modification.
     */
    execute() {
        this.simplifyExpressions();
    }
    /**
     * Simplifies all binary and unary expressions.
     */
    simplifyExpressions() {
        const self = this;
        (0, traverse_1.traverse)(this.ast, {
            enter(node, parent) {
                if (self.types.has(node.type)) {
                    const replacement = self.simplifyExpression(node);
                    if (replacement != node) {
                        traversalHelper_1.default.replaceNode(parent, node, replacement);
                    }
                }
            }
        });
    }
    /**
     * Attempts to simplify an expression node.
     * @param expression The expression node.
     */
    simplifyExpression(expression) {
        switch (expression.type) {
            case 'BinaryExpression':
                return this.simplifyBinaryExpression(expression) || expression;
            case 'UnaryExpression':
                return this.simplifyUnaryExpression(expression) || expression;
            default:
                return expression;
        }
    }
    /**
     * Attempts to simplify a unary expression node.
     * @param expression The unary expression node.
     */
    simplifyUnaryExpression(expression) {
        if (!ExpressionSimplifier.RESOLVABLE_UNARY_OPERATORS.has(expression.operator)) {
            return expression;
        }
        else if (expression.operator == '-' && expression.operand.type == 'LiteralNumericExpression') {
            return expression; // avoid trying to simplify negative numbers
        }
        const argument = this.simplifyExpression(expression.operand);
        if (this.isResolvableExpression(argument)) {
            const argumentValue = this.getResolvableExpressionValue(argument);
            const value = this.applyUnaryOperation(expression.operator, argumentValue);
            return this.convertValueToExpression(value);
        }
        else {
            return expression;
        }
    }
    /**
     * Attempts to simplify a binary expression node.
     * @param expression The binary expression node.
     */
    simplifyBinaryExpression(expression) {
        if (!expression.left.type.endsWith('Expression') ||
            !ExpressionSimplifier.RESOLVABLE_BINARY_OPERATORS.has(expression.operator)) {
            return undefined;
        }
        const left = this.simplifyExpression(expression.left);
        const right = this.simplifyExpression(expression.right);
        if (this.isResolvableExpression(left) && this.isResolvableExpression(right)) {
            const leftValue = this.getResolvableExpressionValue(left);
            const rightValue = this.getResolvableExpressionValue(right);
            const value = this.applyBinaryOperation(expression.operator, leftValue, rightValue);
            return this.convertValueToExpression(value);
        }
        else if (expression.operator == '-' && right.type == 'UnaryExpression' && right.operator == '-' && right.operand.type == 'LiteralNumericExpression') {
            // convert (- -a) to +a (as long as a is a number)
            expression.right = right.operand;
            expression.operator = '+';
            return expression;
        }
        else {
            return undefined;
        }
    }
    /**
     * Applies a unary operation.
     * @param operator The operator.
     * @param argument The argument value.
     * @returns The resultant value.
     */
    applyUnaryOperation(operator, argument) {
        switch (operator) {
            case '-':
                return -argument;
            case '+':
                return +argument;
            case '!':
                return !argument;
            case '~':
                return ~argument;
            case 'typeof':
                return typeof argument;
            case 'void':
                return void argument;
        }
    }
    /**
     * Applies a binary operation.
     * @param operator The resolvable binary operator.
     * @param left The value of the left expression.
     * @param right The value of the right expression.
     * @returns The resultant value.
     */
    applyBinaryOperation(operator, left, right) {
        switch (operator) {
            case '==':
                return left == right;
            case '!=':
                return left != right;
            case '===':
                return left === right;
            case '!==':
                return left !== right;
            case '<':
                return left < right;
            case '<=':
                return left <= right;
            case '>':
                return left > right;
            case '>=':
                return left >= right;
            case '<<':
                return left << right;
            case '>>':
                return left >> right;
            case '>>>':
                return left >>> right;
            case '+':
                return left + right;
            case '-':
                return left - right;
            case '*':
                return left * right;
            case '/':
                return left / right;
            case '%':
                return left % right;
            case '**':
                return left ** right;
            case '|':
                return left | right;
            case '^':
                return left ^ right;
            case '&':
                return left & right;
        }
    }
    /**
     * Gets the real value from a resolvable expression.
     * @param expression The resolvable expression.
     * @returns The value.
     */
    getResolvableExpressionValue(expression) {
        switch (expression.type) {
            case 'LiteralNumericExpression':
            case 'LiteralStringExpression':
            case 'LiteralBooleanExpression':
                return expression.value;
            case 'UnaryExpression':
                return -this.getResolvableExpressionValue(expression.operand);
            case 'LiteralNullExpression':
                return null;
            case 'IdentifierExpression':
                return undefined;
            case 'ArrayExpression':
                return [];
            case 'ObjectExpression':
                return {};
        }
    }
    /**
     * Attempts to convert a value of unknown type to an expression node.
     * @param value The value.
     * @returns The expression or undefined.
     */
    convertValueToExpression(value) {
        switch (typeof value) {
            case 'string':
                return new Shift.LiteralStringExpression({ value });
            case 'number':
                return value >= 0
                    ? new Shift.LiteralNumericExpression({ value })
                    : new Shift.UnaryExpression({ operator: '-', operand: new Shift.LiteralNumericExpression({ value: Math.abs(value) }) });
            case 'boolean':
                return new Shift.LiteralBooleanExpression({ value });
            case 'undefined':
                return new Shift.IdentifierExpression({ name: 'undefined' });
            default:
                return undefined;
        }
    }
    /**
     * Returns whether a node is a resolvable expression that can be
     * evaluated safely.
     * @param node The AST node.
     * @returns Whether.
     */
    isResolvableExpression(node) {
        return (this.isLiteral(node) ||
            (node.type == 'UnaryExpression' && node.operator == '-' && node.operand.type == 'LiteralNumericExpression') ||
            (node.type == 'IdentifierExpression' && node.name == 'undefined') ||
            (node.type == 'ArrayExpression' && node.elements.length == 0) ||
            (node.type == 'ObjectExpression' && node.properties.length == 0));
    }
    /**
     * Returns whether a node is a literal.
     * @param node The AST node.
     * @returns Whether.
     */
    isLiteral(node) {
        return node.type == 'LiteralNumericExpression' || node.type == 'LiteralStringExpression' || node.type == 'LiteralBooleanExpression' || node.type == 'LiteralNullExpression';
    }
}
exports.default = ExpressionSimplifier;
ExpressionSimplifier.RESOLVABLE_UNARY_OPERATORS = new Set([
    '-',
    '+',
    '!',
    '~',
    'typeof',
    'void'
]);
ExpressionSimplifier.RESOLVABLE_BINARY_OPERATORS = new Set([
    '==',
    '!=',
    '===',
    '!==',
    '<',
    '<=',
    '>',
    '>=',
    '<<',
    '>>',
    '>>>',
    '+',
    '-',
    '*',
    '/',
    '%',
    '**',
    '|',
    '^',
    '&'
]);
