"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ScopeType = void 0;
class Scope {
    /**
     * Creates a new scope.
     * @param node The node that created the scope.
     * @param type The type of scope.
     * @param parent The parent scope (optional).
     */
    constructor(node, type, parent) {
        this.node = node;
        this.type = type;
        this.parent = parent;
        this.children = new Map();
        this.elements = new Map();
        if (this.parent) {
            this.parent.children.set(this.node, this);
        }
    }
    /**
     * Gets an element by name.
     * @param name The name associated with the element.
     * @returns The element or null.
     */
    get(name) {
        if (this.elements.has(name)) {
            return this.elements.get(name);
        }
        return this.parent ? this.parent.get(name) : null;
    }
    /**
     * Adds an element.
     * @param name The name associated with the element.
     * @param element The element.
     * @param type The type of the variable.
     */
    add(name, element, type = 'const') {
        switch (type) {
            case 'const':
            case 'let': {
                this.elements.set(name, element);
                break;
            }
            case 'var': {
                const scope = this.findScope([ScopeType.Function, ScopeType.Global]);
                if (!scope) {
                    throw new Error(`Failed to find scope for var ${name}`);
                }
                scope.elements.set(name, element);
                break;
            }
            case undefined: {
                const scope = this.findScope([ScopeType.Global]);
                if (!scope) {
                    throw new Error(`Failed to find scope for global var ${name}`);
                }
                scope.elements.set(name, element);
                break;
            }
        }
    }
    /**
     * Gets the scope for a given declaration type.
     * @param type The declaration type.
     * @returns The scope.
     */
    getDeclarationScope(type) {
        switch (type) {
            case 'const':
            case 'let':
                return this;
            case 'var':
            case 'global': {
                const scopeTypes = type === 'var' ? [ScopeType.Function, ScopeType.Global] : [ScopeType.Global];
                const scope = this.findScope(scopeTypes);
                if (!scope) {
                    throw new Error(`Failed to find scope for variable declaration type ${type}`);
                }
                return scope;
            }
        }
    }
    /**
     * Finds a scope with a given type. Looks back up the scope tree.
     * @param types The desired scope types.
     * @returns The scope found or undefined.
     */
    findScope(types) {
        let scope = this;
        while (scope && !types.includes(scope.type)) {
            scope = scope.parent;
        }
        return scope;
    }
}
exports.default = Scope;
var ScopeType;
(function (ScopeType) {
    ScopeType["Global"] = "Global";
    ScopeType["Function"] = "Function";
    ScopeType["Other"] = "Other";
})(ScopeType = exports.ScopeType || (exports.ScopeType = {}));
